<?php

class Matrix
{
    private $M; // this matrix as a 2-D array
    private $Tol;       // used in LUPDecompose
    // LUP Decomposition, calculated on deamand (before first use of LUPSolve, LUPInvert, LUPDeterminant)
    private $A =   null; // LU decomposition of the matrix
    private $P =   null; // permutation array as defined in Wikipedia code:
    public function __construct($M, $Tol = 1E-6){
//        print("__construct\n");
        if (is_array($M)){
            $this->M = $M;
            if (!is_array($M[0])){ // single-column array
                for ($i = 0; $i< sizeof($M); $i++){
                    $this->M[$i] = array($M[$i]);
                }
            }
        }
        $this->Tol = $Tol;
    }
    public function get(){
        return $this->M;
    }
    
    public function set($i,$j,$v){
        $this->M[$i][$j] = $v;
    }
    
    
    public function getColumnPackedCopy(){
        $rows = sizeof($this->M);
        $cols = sizeof($this->M[0]);
        $R = array_fill(0, $rows * $cols, 0.0);
        $indx = 0;
        for ($i = 0; $i < $rows; $i++){
            for ($j = 0; $j < $cols; $j++){
                $R[$indx++] = $this->M[$i][$j];
            }
        }
        return $R;
    }
    
    public function getRowPackedCopy(){
        $rows = sizeof($this->M);
        $cols = sizeof($this->M[0]);
        $R = array_fill(0, $rows * $cols, 0.0);
        $indx = 0;
//        var_dump($this->M);
        for ($j = 0; $j < $cols; $j++){
            for ($i = 0; $i < $rows; $i++){
                $R[$indx++] = $this->M[$i][$j];
            }
        }
        return $R;
        
    }
    public function times($B){
        if ($B instanceof Matrix){
            $B = $B->get();
        }
        if (is_array($B)) {
            $rows1 = sizeof($this->M);
            $cols1 = sizeof($this->M[0]);
            if ($cols1 != sizeof($B)){
                throw new Exception('Dimensions mismatch.');
            }
            $cols2 = sizeof($B[0]);
            $R = $this->zeroMatrix($rows1, $cols2);
            for ($i = 0; $i < $rows1; $i++){
                for ($j = 0; $j < $cols2; $j++){
                    for ($k = 0; $k < $cols1; $k++){
                        $R[$i][$j] += $this->M[$i][$k] * $B[$k][$j];
                    }
                }
            }
        } else { // times scalar
            $R=$this->M;
            $rows = sizeof($this->M);
            $cols = sizeof($this->M[0]);
            for ($i = 0; $i < $rows; $i++){
                for ($j = 0; $j < $cols; $j++){
                    $R[$i][$j] *= $B;
                }
            }
        }
        return new Matrix ($R, $this->Tol);
    }

    public function plus($B) {
        if ($B instanceof Matrix){
            $B = $B->get();
        }
        $R = $this->M;
        $rows = sizeof($this->M);
        $cols = sizeof($this->M[0]);
        if (($rows != sizeof($B[0])) || ($cols != sizeof($B[0]))){
            throw new Exception('Dimensions mismatch.');
        }
        for ($i = 0; $i < $rows; $i ++) {
            for ($j = 0; $j < $cols; $j ++) {
                $R[$i][$j] += $B[$i][$j];
            }
        }
        return new Matrix ($R, $this->Tol);
    }
    public function plusEquals($B){
        if ($B instanceof Matrix){
            $B = $B->get();
        }
        $rows = sizeof($this->M);
        $cols = sizeof($this->M[0]);
        if (($rows != sizeof($B[0])) || ($cols != sizeof($B[0]))){
            throw new Exception('Dimensions mismatch.');
        }
        for ($i = 0; $i < $rows; $i ++) {
            for ($j = 0; $j < $cols; $j ++) {
                $this->M[$i][$j] += $B[$i][$j];
            }
        }
    }
    
    
    public function print($debugFile = null, $decimals = 6)
    {
        if ($debugFile === null){
            $debugFile = fopen('php://stdout', 'w');
        }
        foreach ($this->M as $row) {
            fprintf($debugFile, "[");
            foreach ($row as $i) {
                fprintf($debugFile, "\t" . sprintf("%01.{$decimals}f", round($i, $decimals)));
            }
            fprintf($debugFile,  "\t]\n");
        }
    }
    
    
    /*
     *        The permutation matrix is not stored as a matrix, but in an integer vector P of size N+1
     *        containing column indexes where the permutation matrix has "1". The last element P[N]=S+N,
     *        where S is the number of row exchanges needed for determinant computation, det(P)=(-1)^S
     */
    // Ported from C code in Wikipedia https://en.wikipedia.org/wiki/LU_decomposition
    /* INPUT: A - array of pointers to rows of a square matrix having dimension N
     *        Tol - small tolerance number to detect failure when the matrix is near degenerate
     * OUTPUT: Matrix A is changed, it contains a copy of both matrices L-E and U as A=(L-E)+U such that P*A=L*U.
     *        The permutation matrix is not stored as a matrix, but in an integer vector P of size N+1
     *        containing column indexes where the permutation matrix has "1". The last element P[N]=S+N,
     *        where S is the number of row exchanges needed for determinant computation, det(P)=(-1)^S
     */
    /**
     * Perform LU decomposition
     * @param $A - square 2D array as input matrix, see Wikipedia comments above 
     * @param $Tol - tolerance to catch singilar matrix
     * @return - array $P for permutation matrix on success, null on failure (singular)
     */

    public function LUPDecompose (){
        $N = sizeof($this->M);
        $this->A = $this->M;
        $this->P = array_fill(0,$N+1, 0.0);
        for ($i = 0; $i <= $N; $i++) $this->P[$i] = $i; //Unit permutation matrix, P[N] initialized with N
        for ($i = 0; $i < $N; $i++) {
            $maxA = 0.0;
            $imax = $i;
            for ($k = $i; $k < $N; $k++) {
                $absA = abs($this->A[$k][$i]);
                if ($absA > $maxA) {
                    $maxA = $absA;
                    $imax = $k;
                }
            }
            if ($maxA < $this->Tol) {
                throw new Exception('Matrix is degenerate.'); //failure, matrix is degenerate
            }
            
            if ($imax != $i) {
                //pivoting P
                $j = $this->P[$i];
                $this->P[$i] = $this->P[$imax];
                $this->P[$imax] = $j;
                
                //pivoting rows of A
                $ptr =   $this->A[$i];
                $this->A[$i] = $this->A[$imax];
                $this->A[$imax] = $ptr;
                
                //counting pivots starting from N (for determinant)
                $this->P[$N]++;
            }
            
            for ($j = $i + 1; $j < $N; $j++) {
                $this->A[$j][$i] /= $this->A[$i][$i];
                
                for ($k = $i + 1; $k < $N; $k++)
                    $this->A[$j][$k] -= $this->A[$j][$i] * $this->A[$i][$k];
            }
        }
    }
    /* INPUT: A,P filled in LUPDecompose; b - rhs vector; N - dimension
     * OUTPUT: x - solution vector of A*x=b
     */
    public function solve($b) {
        if ($this->A === null){
            $this->LUPDecompose();
        }
        if ($b instanceof Matrix){
            $b = $b->getColumnPackedCopy();
        }
        $N = sizeof($this->M);
        $x = array_fill(0,$N,0.0);
        for ($i = 0; $i < $N; $i++) {
            $x[$i] = $b[$this->P[$i]];
            for ($k = 0; $k < $i; $k++)
                $x[$i] -= $this->A[$i][$k] * $x[$k];
        }
        for ($i = $N - 1; $i >= 0; $i--) {
            for ($k = $i + 1; $k < $N; $k++)
                $x[$i] -= $this->A[$i][$k] * $x[$k];
                $x[$i] = $x[$i] / $this->A[$i][$i];
        }
        return new Matrix ($x, $this->Tol);
    }

        /*
     * INPUT: A,P filled in LUPDecompose; N - dimension
     * OUTPUT: IA is the inverse of the initial matrix
     */
    public function invert()
    {
        if ($this->A === null){
            $this->LUPDecompose();
        }
        $N = sizeof($this->A);
        $IA = $this->zeroMatrix($N, $N);
        for ($j = 0; $j < $N; $j ++) {
            for ($i = 0; $i < $N; $i ++) {
                if ($this->P[$i] == $j)
                    $IA[$i][$j] = 1.0;
                else
                    $IA[$i][$j] = 0.0;
                for ($k = 0; $k < $i; $k ++)
                    $IA[$i][$j] -= $this->A[$i][$k] * $IA[$k][$j];
            }
            for ($i = $N - 1; $i >= 0; $i --) {
                for ($k = $i + 1; $k < $N; $k ++)
                    $IA[$i][$j] -= $this->A[$i][$k] * $IA[$k][$j];
                $IA[$i][$j] = $IA[$i][$j] / $this->A[$i][$i];
            }
        }
        return new Matrix ($IA, $this->Tol);
    }

     /*
     * INPUT: A,P filled in LUPDecompose; N - dimension.
     * OUTPUT: Function returns the determinant of the initial matrix
     */
    public function det()
    {
        if ($this->A === null){
            $this->LUPDecompose();
        }
        $N = sizeof($this->A);
        $det = $this->A[0][0];
        for ($i = 1; $i < $N; $i ++)
            $det *= $this->A[$i][$i];
            if (($this->P[$N] - $N) % 2 == 0)
            return $det;
        else
            return - $det;
    }
    
    public function isNonsingular(){
        if ($this->A === null){
            try {
            $this->LUPDecompose();
            } catch (Exception $e){
                return 0;  // Singular
            }
        }
        return ($this->det())!=0.0; // Non-singular
    }
    
    public static function zeroMatrix($m,$n=1){
        $M = array_fill(0, $m, null);
        for ($i = 0; $i < $m; $i++){
            $M[$i] = array_fill(0, $n, 0.0);
        }
        return $M;
    }
    
    public static function identityMatrix($n)
    {
        $I = array();
        for ($i = 0; $i < $n; ++ $i) {
            for ($j = 0; $j < $n; ++ $j) {
                $I[$i][$j] = ($i == $j) ? 1 : 0;
            }
        }
        return $I;
    }
    
}
/*
// test code here
$A = array(
    array( 10, -15,  30,   6, -8 ),
    array(  0,  -4,  60,  11, -5 ),
    array(  8,   9,   2,   3,  7 ),
    array( 25,  10,  -9,   9,  0 ),
    array( 13,   3, -12,   5,  2 ),
);
$SING = array(
    array( 10, -15,  30,   6, -8 ),
    array(  0,  -4,  60,  11, -5 ),
    array(  8,   9,   2,   3,  7 ),
    array( 25,  10,  -9,   9,  0 ),
    array( 10, -15,  30,   6, -8 ),
);


$M1 = new Matrix($A);
print("\nM1=\n");
$M1->print(); 
//exit(0);
$IM1 = $M1->invert();
print("\nIM1=\n");
$IM1->print();
$M2 = $IM1->invert();
print("\nM2=\n");
$M2->print();

$det= $M1->det();
print("\ndet(M1)=".$det."\n");
$idet= $IM1->det();
print("\ndet(IM1)=".$idet."\n");
print("\ndet(M1)*det(IM1)=".($det * $idet)."\n");

$MS = new Matrix($SING);
print("\nM1 nonsingular=".$M1->isNonsingular()."\n");
print("\nMS nonsingular=".$MS->isNonsingular()."\n");

print("\nM1 * IM1=\n");
$M1->times($IM1)->print();
*/
?>
